// Copyright 2021 The Oppia Authors. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS-IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

/**
 * @fileoverview Unit tests for supplemental card component.
 */

import { HttpClientTestingModule } from '@angular/common/http/testing';
import { ChangeDetectorRef, EventEmitter, NO_ERRORS_SCHEMA, SimpleChanges } from '@angular/core';
import { ComponentFixture, fakeAsync, TestBed, tick, waitForAsync } from '@angular/core/testing';
import { AppConstants } from 'app.constants';
import { StateCard } from 'domain/state_card/state-card.model';
import { UrlInterpolationService } from 'domain/utilities/url-interpolation.service';
import { AudioPlayerService } from 'services/audio-player.service';
import { AutogeneratedAudioPlayerService } from 'services/autogenerated-audio-player.service';
import { WindowRef } from 'services/contextual/window-ref.service';
import { ExplorationPlayerConstants } from '../exploration-player-page.constants';
import { AudioTranslationManagerService } from '../services/audio-translation-manager.service';
import { CurrentInteractionService } from '../services/current-interaction.service';
import { HelpCardEventResponse, PlayerPositionService } from '../services/player-position.service';
import { SupplementalCardComponent } from './supplemental-card.component';
import { I18nLanguageCodeService } from 'services/i18n-language-code.service';
import { Interaction } from 'domain/exploration/InteractionObjectFactory';
import { RecordedVoiceovers } from 'domain/exploration/recorded-voiceovers.model';
import { AudioTranslationLanguageService } from '../services/audio-translation-language.service';

describe('Supplemental card component', () => {
  let fixture: ComponentFixture<SupplementalCardComponent>;
  let componentInstance: SupplementalCardComponent;
  let audioPlayerService: AudioPlayerService;
  let audioTranslationManagerService: AudioTranslationManagerService;
  let autogeneratedAudioPlayerService: AutogeneratedAudioPlayerService;
  let currentInteractionService: CurrentInteractionService;
  let playerPositionService: PlayerPositionService;
  let urlInterpolationService: UrlInterpolationService;
  let windowRef: WindowRef;
  let i18nLanguageCodeService: I18nLanguageCodeService;
  let recordedVoiceovers = new RecordedVoiceovers({});
  let mockStateCard = new StateCard(
    'state_name', 'html', 'html', {} as Interaction, [], recordedVoiceovers,
    '', {} as AudioTranslationLanguageService);

  class MockChangeDetectorRef {
    detectChanges(): void {}
  }

  beforeEach(waitForAsync(() => {
    TestBed.configureTestingModule({
      imports: [
        HttpClientTestingModule
      ],
      declarations: [
        SupplementalCardComponent
      ],
      providers: [
        AudioPlayerService,
        AudioTranslationManagerService,
        AutogeneratedAudioPlayerService,
        {
          provide: ChangeDetectorRef,
          useClass: MockChangeDetectorRef
        },
        CurrentInteractionService,
        PlayerPositionService,
        UrlInterpolationService,
        WindowRef
      ],
      schemas: [NO_ERRORS_SCHEMA]
    }).compileComponents();
  }));

  beforeEach(() => {
    fixture = TestBed.createComponent(SupplementalCardComponent);
    componentInstance = fixture.componentInstance;
    audioPlayerService = TestBed.inject(AudioPlayerService);
    audioTranslationManagerService = TestBed.inject(
      AudioTranslationManagerService);
    autogeneratedAudioPlayerService = TestBed.inject(
      AutogeneratedAudioPlayerService);
    currentInteractionService = TestBed.inject(CurrentInteractionService);
    playerPositionService = TestBed.inject(PlayerPositionService);
    urlInterpolationService = TestBed.inject(UrlInterpolationService);
    windowRef = TestBed.inject(WindowRef);
    i18nLanguageCodeService = TestBed.inject(I18nLanguageCodeService);

    spyOn(i18nLanguageCodeService, 'isCurrentLanguageRTL').and.returnValue(
      true);
  });

  afterEach(() => {
    componentInstance.ngOnDestroy();
  });

  it('should intialize', fakeAsync(() => {
    let imageUrl = 'static_image_url';
    let helpCardHtml = 'test_help_card_html';
    let hasContinueButton = true;
    componentInstance.currentDisplayedCard = mockStateCard;
    spyOn(urlInterpolationService, 'getStaticImageUrl').and.returnValue(
      imageUrl);
    spyOn(currentInteractionService, 'registerPresubmitHook').and.callFake(
      callb => {
        callb();
      });
    spyOn(componentInstance.currentDisplayedCard, 'isCompleted')
      .and.returnValues(false, true);
    spyOn(componentInstance, 'clearHelpCard');

    let mockOnActiveCardChanged = new EventEmitter<void>();
    let mockOnHelpCardAvailable = new EventEmitter<HelpCardEventResponse>();

    spyOnProperty(playerPositionService, 'onActiveCardChanged')
      .and.returnValue(mockOnActiveCardChanged);
    spyOn(componentInstance, 'updateDisplayedCard');
    spyOnProperty(playerPositionService, 'onHelpCardAvailable').and.returnValue(
      mockOnHelpCardAvailable);

    componentInstance.ngOnInit();

    mockOnActiveCardChanged.emit();
    mockOnHelpCardAvailable.emit({
      helpCardHtml,
      hasContinueButton
    });
    tick();
    tick();
    componentInstance.ngOnInit();
    expect(componentInstance.helpCardHtml).toEqual(helpCardHtml);
    expect(componentInstance.helpCardHasContinueButton).toEqual(
      hasContinueButton);
    expect(componentInstance.clearHelpCard).toHaveBeenCalled();
    expect(componentInstance.updateDisplayedCard).toHaveBeenCalled();
  }));

  it('should update displayed card', () => {
    let lastAnswer = 'last_answer';
    mockStateCard.markAsCompleted();
    componentInstance.displayedCard = mockStateCard;
    spyOn(componentInstance, 'clearHelpCard');
    spyOn(componentInstance.displayedCard, 'getLastAnswer')
      .and.returnValue(lastAnswer);
    componentInstance.updateDisplayedCard();
    expect(componentInstance.lastAnswer).toEqual(lastAnswer);
  });

  it('should clear help card', () => {
    componentInstance.clearHelpCard();
    expect(componentInstance.helpCardHtml).toBeNull();
    expect(componentInstance.helpCardHasContinueButton).toBeFalse();
    expect(componentInstance.maxHelpCardHeightSeen).toEqual(0);
  });

  it('should tell if help card is tall', () => {
    let height = 400;
    spyOn(
      fixture.elementRef.nativeElement, 'getElementsByClassName'
    ).withArgs('conversation-skin-help-card').and.returnValue([
      {
        offsetHeight: 400
      }
    ]);
    spyOn(componentInstance, 'updateHelpCardBottomPosition');
    componentInstance.maxHelpCardHeightSeen = height - 100;
    expect(componentInstance.isHelpCardTall()).toBeFalse();
    spyOnProperty(windowRef.nativeWindow, 'innerHeight').and.returnValue(100);
    expect(componentInstance.isHelpCardTall()).toBeTrue();
  });

  it('should update help card bottom position', () => {
    let helpCardHeight = 100;
    let interactionContainerHeight = 400;
    componentInstance.currentDisplayedCard = mockStateCard;
    componentInstance.helpCard = {
      nativeElement: {
        clientHeight: helpCardHeight
      }
    };
    componentInstance.interactionContainer = {
      nativeElement: {
        clientHeight: interactionContainerHeight
      }
    };
    componentInstance.updateHelpCardBottomPosition();

    expect(componentInstance.helpCardBottomPosition).toEqual(350);
  });

  it('should get feedback audio highlight class', () => {
    spyOn(
      audioTranslationManagerService, 'getCurrentComponentName')
      .and.returnValue(AppConstants.COMPONENT_NAME_FEEDBACK);
    spyOn(audioPlayerService, 'isPlaying').and.returnValue(true);
    spyOn(autogeneratedAudioPlayerService, 'isPlaying').and.returnValue(true);
    expect(componentInstance.getFeedbackAudioHighlightClass()).toEqual(
      ExplorationPlayerConstants.AUDIO_HIGHLIGHT_CSS_CLASS);
  });

  it('should return null if audio player is not playing', () => {
    spyOn(
      audioTranslationManagerService, 'getCurrentComponentName')
      .and.returnValue(AppConstants.COMPONENT_NAME_FEEDBACK);
    spyOn(audioPlayerService, 'isPlaying').and.returnValue(false);
    spyOn(autogeneratedAudioPlayerService, 'isPlaying').and.returnValue(false);
    expect(componentInstance.getFeedbackAudioHighlightClass()).toEqual(null);
  });

  it('should update the displayedCard when changes are detected', () => {
    let updatedStateCard = new StateCard(
      'state_name', 'new content', 'html', {} as Interaction, [],
      recordedVoiceovers, '',
      {} as AudioTranslationLanguageService);
    const changes: SimpleChanges = {
      displayedCard: {
        previousValue: mockStateCard,
        currentValue: updatedStateCard,
        firstChange: false,
        isFirstChange: () => false
      }
    };
    componentInstance.displayedCard = mockStateCard;
    expect(componentInstance.displayedCard.contentHtml).toBe('html');

    componentInstance.ngOnChanges(changes);

    expect(componentInstance.displayedCard.contentHtml).toBe('new content');
  });
});
