// Copyright 2021 The Oppia Authors. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS-IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

/**
 * @fileoverview Component for display hint modal.
 */

import { Component } from '@angular/core';
import { NgbActiveModal } from '@ng-bootstrap/ng-bootstrap';
import { RecordedVoiceovers } from 'domain/exploration/recorded-voiceovers.model';
import { SubtitledHtml } from 'domain/exploration/subtitled-html.model';
import { StateCard } from 'domain/state_card/state-card.model';
import { AudioPlayerService } from 'services/audio-player.service';
import { AutogeneratedAudioPlayerService } from 'services/autogenerated-audio-player.service';
import { AudioTranslationManagerService } from '../services/audio-translation-manager.service';
import { HintsAndSolutionManagerService } from '../services/hints-and-solution-manager.service';
import { PlayerPositionService } from '../services/player-position.service';
import { PlayerTranscriptService } from '../services/player-transcript.service';

@Component({
  selector: 'oppia-display-hint-modal',
  templateUrl: './display-hint-modal.component.html'
})
export class DisplayHintModalComponent {
  // These properties are initialized using Angular lifecycle hooks
  // and we need to do non-null assertion. For more information, see
  // https://github.com/oppia/oppia/wiki/Guide-on-defining-types#ts-7-1
  index!: number;
  COMPONENT_NAME_HINT!: string;
  hint!: SubtitledHtml;
  displayedCard!: StateCard;
  recordedVoiceovers!: RecordedVoiceovers;
  hintContentId!: string;

  constructor(
    private ngbActiveModal: NgbActiveModal,
    private audioPlayerService: AudioPlayerService,
    private audioTranslationManagerService: AudioTranslationManagerService,
    private autogeneratedAudioPlayerService: AutogeneratedAudioPlayerService,
    private hintsAndSolutionManagerService: HintsAndSolutionManagerService,
    private playerPositionService: PlayerPositionService,
    private playerTranscriptService: PlayerTranscriptService
  ) {}

  ngOnInit(): void {
    let displayHint = (
      this.hintsAndSolutionManagerService.displayHint(this.index));
    if (displayHint === null) {
      throw new Error('Hint not found.');
    }
    this.hint = displayHint;
    this.displayedCard = this.playerTranscriptService.getCard(
      this.playerPositionService.getDisplayedCardIndex());
    this.recordedVoiceovers = this.displayedCard.getRecordedVoiceovers();
    let contentId = this.hint.contentId;
    if (contentId === null) {
      throw new Error('Content id not found.');
    }
    this.hintContentId = contentId;

    this.audioTranslationManagerService
      .setSecondaryAudioTranslations(
        this.recordedVoiceovers.getBindableVoiceovers(this.hintContentId),
        this.hint.html, this.COMPONENT_NAME_HINT);

    this.audioPlayerService.onAutoplayAudio.emit();
  }

  closeModal(): void {
    this.audioPlayerService.stop();
    this.autogeneratedAudioPlayerService.cancel();
    this.audioTranslationManagerService.clearSecondaryAudioTranslations();
    this.ngbActiveModal.dismiss('cancel');
  }
}
